"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.associate_route_table
hub.exec.boto3.client.ec2.create_route_table
hub.exec.boto3.client.ec2.delete_route_table
hub.exec.boto3.client.ec2.describe_route_tables
hub.exec.boto3.client.ec2.disassociate_route_table
resource = hub.tool.boto3.resource.create(ctx, "ec2", "RouteTable", name)
hub.tool.boto3.resource.exec(resource, associate_with_subnet, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_route, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_tags, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    vpc_id: str,
    resource_id: str = None,
    routes: List[
        make_dataclass(
            "Route",
            [
                ("DestinationCidrBlock", str, field(default=None)),
                ("DestinationIpv6CidrBlock", str, field(default=None)),
                ("DestinationPrefixListId", str, field(default=None)),
                ("EgressOnlyInternetGatewayId", str, field(default=None)),
                ("GatewayId", str, field(default=None)),
                ("InstanceId", str, field(default=None)),
                ("InstanceOwnerId", str, field(default=None)),
                ("NatGatewayId", str, field(default=None)),
                ("TransitGatewayId", str, field(default=None)),
                ("LocalGatewayId", str, field(default=None)),
                ("CarrierGatewayId", str, field(default=None)),
                ("NetworkInterfaceId", str, field(default=None)),
                ("VpcPeeringConnectionId", str, field(default=None)),
                ("CoreNetworkArn", str, field(default=None)),
            ],
        )
    ] = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a route table for the specified VPC. For more information, see Route tables in the Amazon Virtual Private Cloud User Guide.

    Args:
        name(Text): An Idem name to identify the route table resource.
        resource_id(Text, optional): AWS Route Table ID.
        vpc_id(Text): AWS VPC ID.
        routes(List[Dict[Text, Any]], optional): The network interfaces to associate with the instance. If you specify a network interface, you
            must specify any security groups and subnets as part of the network interface. Defaults to None.
            * DestinationCidrBlock (str, optional): The IPv4 CIDR address block used for the destination match. Routing decisions are based on the
                most specific match. We modify the specified CIDR block to its canonical form; for example, if
                you specify 100.68.0.18/18, we modify it to 100.68.0.0/18. Defaults to None.
            * DestinationIpv6CidrBlock (str, optional): The IPv6 CIDR block used for the destination match.
                Routing decisions are based on the most specific match. Defaults to None.
            * EgressOnlyInternetGatewayId (str, optional): [IPv6 traffic only] The ID of an egress-only internet gateway. Defaults to None.
            * GatewayId (str, optional): The ID of an internet gateway or virtual private gateway attached to your VPC. Defaults to None.
            * InstanceId (str, optional): The ID of a NAT instance in your VPC. The operation fails if you specify an instance ID unless
                exactly one network interface is attached. Defaults to None.
            * InstanceOwnerId (str, optional): The ID of Amazon Web Services account that owns the instance. Defaults to None.
            * NatGatewayId (str, optional): [IPv4 traffic only] The ID of a NAT gateway. Defaults to None.
            * (TransitGatewayId, str, optional): The ID of a transit gateway. Defaults to None.
            * LocalGatewayId (str, optional): The ID of the local gateway. Defaults to None.
            * CarrierGatewayId (str, optional): The ID of the carrier gateway. You can only use this option when the VPC contains a subnet which
                is associated with a Wavelength Zone. Defaults to None.
            * NetworkInterfaceId (str, optional): The ID of a network interface. Defaults to None.
            * VpcPeeringConnectionId (str, optional): The ID of a VPC peering connection. Defaults to None.
            * CoreNetworkArn (str, optional): The Amazon Resource Name (ARN) of the core network. Defaults to None.
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the route table.
            Each tag consists of a key name and an associated value. Defaults to None.
            * Key (str, optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                characters. May not begin with aws:.
            * Value (str, optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.

    Request Syntax:
        [route_table-resource-name]:
          aws.ec2.route_table.present:
          - vpc_id: 'string'
          - resource_id: 'string'
          - routes:
            - DestinationCidrBlock: 'string'
              GatewayId: 'string'
          - tags:
            - Key: 'string'
              Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            test_route-table:
              aws.ec2.route_table.present:
              - vpc_id: vpc-02850adfa9f6fc916
              - resource_id: route_table-3485hydfe5f6tb998
              - routes:
                - DestinationCidrBlock: 198.31.0.0/16
                  GatewayId: local
              - tags:
                - Key: Name
                  Value: route-table-association-test
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    update_successful = False
    plan_state = {}
    before = None
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
    if resource_id:
        try:
            before = await hub.exec.aws.ec2.route_table.get(
                ctx=ctx, name=name, resource_id=resource_id
            )
            if not before["result"] or not before["ret"]:
                result["result"] = False
                result["comment"] = before["comment"]
                return result
            result["old_state"] = copy.deepcopy(before["ret"])
            plan_state = copy.deepcopy(result["old_state"])
            update_ret = await hub.exec.aws.ec2.route_table.update_routes(
                ctx=ctx,
                route_table_id=resource_id,
                old_routes=result["old_state"].get("routes"),
                new_routes=routes,
            )
            result["comment"] = result["comment"] + tuple(update_ret["comment"])
            update_successful = update_ret["result"]
            if not update_ret["result"]:
                result["result"] = False
            if update_ret["ret"] and ctx.get("test", False):
                if update_ret["ret"].get("routes") is not None:
                    plan_state["routes"] = update_ret["ret"].get("routes")
            if (
                update_successful
                and tags is not None
                and tags != result["old_state"].get("tags")
            ):
                # Update tags
                update_ret = await hub.exec.aws.ec2.tag.update_tags(
                    ctx=ctx,
                    resource_id=resource_id,
                    old_tags=result["old_state"].get("tags"),
                    new_tags=tags,
                )
                update_successful = update_ret["result"]
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = update_ret["result"]
                if ctx.get("test", False) and update_ret["result"]:
                    plan_state["tags"] = update_ret["ret"]

        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "vpc_id": vpc_id,
                    "resource_id": name,
                    "routes": routes,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.ec2.route_table", name=name
            )
            return result
        try:
            ret = await hub.exec.boto3.client.ec2.create_route_table(
                ctx,
                TagSpecifications=[
                    {
                        "ResourceType": "route-table",
                        "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
                    }
                ]
                if tags
                else None,
                VpcId=vpc_id,
            )
            result["result"] = ret["ret"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            resource_id = ret["ret"]["RouteTable"]["RouteTableId"]
            result["comment"] = result["comment"] + (f"Created '{name}'",)
            # Associate routes if provided
            update_ret = await hub.exec.aws.ec2.route_table.update_routes(
                ctx=ctx,
                route_table_id=resource_id,
                old_routes=[],
                new_routes=routes,
            )
            result["comment"] = result["comment"] + tuple(update_ret["comment"])
            result["result"] = result["result"] and update_ret["result"]
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif update_successful or not (before and before["result"]):
            after = await hub.exec.aws.ec2.route_table.get(
                ctx=ctx, name=name, resource_id=resource_id
            )
            if not after["result"]:
                result["result"] = False
                result["comment"] = after["comment"]
                return result
            result["new_state"] = copy.deepcopy(after["ret"])
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified route table. You must disassociate the route table from any subnets before you can delete
    it. You can't delete the main route table.

    Args:
        name(Text): An Idem name to identify the route table resource.
        resource_id(Text, optional): AWS Route Table ID. Idem automatically considers this resource being absent
         if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.ec2.route_table.absent:
                - name: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.route_table", name=name
        )
        return result
    before = await hub.exec.aws.ec2.route_table.get(
        ctx=ctx, name=name, resource_id=resource_id
    )
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result
    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.route_table", name=name
        )
    else:
        result["old_state"] = before["ret"]
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.ec2.route_table", name=name
            )
            return result
        else:
            ret = await hub.exec.boto3.client.ec2.delete_route_table(
                ctx, RouteTableId=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                result["result"] = False
                return result
            result["comment"] = result[
                "comment"
            ] + hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.ec2.route_table", name=name
            )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Describes one or more of your route tables. Each subnet in your VPC must be associated with a route table. If a
    subnet is not explicitly associated with any route table, it is implicitly associated with the main route table.
    This command does not return the subnet ID for implicit associations. For more information, see Route tables in
    the Amazon Virtual Private Cloud User Guide.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ec2.route_table
    """

    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_route_tables(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe route_table {ret['comment']}")
        return result
    for route_table in ret["ret"]["RouteTables"]:
        route_table_id = route_table.get("RouteTableId")
        resource_converted = (
            hub.tool.aws.ec2.conversion_utils.convert_raw_route_table_to_present(
                raw_resource=route_table,
                idem_resource_name=route_table_id,
            )
        )
        result[route_table_id] = {
            "aws.ec2.route_table.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_converted.items()
            ]
        }
    return result


async def search(hub, ctx, name, resource_id: str = None, filters: List = None):
    r"""
    Provides details about a specific route table as a data-source. Supply one of the inputs as the filter.

    Args:
        name(Text): An Idem name of the AWS route table resource.
        resource_id(Text, optional): AWS route table ID to identify the resource.
        filters(list, optional): One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
         https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_route_tables

    Request Syntax:
        [Idem-state-name]:
          aws.ec2.route_table.search:
              - name: 'string'
              - resource_id: 'string'
              - filters:
                - name: 'string'
                  values: list
                - name: 'string'
                  values: list

    Response Syntax:
        The following fields will be returned in the `new_state` of the response.

        - name: 'string'
        - resource_id: 'string'
        - vpc_id: 'string'
        - propagating_vgws: list
        - tags: list
        - routes: list
        - associations: list

    Examples:

        Input state file:

        .. code-block:: sls

            idem-test-route-table-search:
                aws.ec2.route_table.search:
                  - name: idem-test-route-table-search
                  - filters:
                    - name: 'vpc-id'
                      values: ["vpc-0ba2072e7101375ee"]
                    - name: 'tag:Name'
                      values: ["cluster01-eks-private-0"]

        Sample response:

            - name: 'idem-test-route-table-search'
            - resource_id: 'rtb-05d27d2c959185162'
            - vpc_id: 'vpc-0ba2072e7101375ee'
            - propagating_vgws: []
            - tags:
                - Key: 'Name'
                  Value: 'cluster01-eks-private-0'
            - routes:
                - DestinationCidrBlock: '10.170.0.0/16'
                  GatewayId: 'local'
                  Origin: 'CreateRouteTable'
                  State: 'active'
            - associations:
                - Main: False
                  RouteTableAssociationId: 'rtbassoc-009e17e6d53ddf889'
                  RouteTableId: 'rtb-05d27d2c959185162'
                  SubnetId: 'subnet-023a9ac1f67decab0'
                  AssociationState:
                    State: 'associated'

    """
    hub.log.warning(
        f"aws.ec2.route_table.search '{name}' state has been deprecated. Please use exec.run with aws.ec2.route_table.get instead."
    )

    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)
    ret = await hub.exec.aws.ec2.route_table.get(
        ctx=ctx,
        name=name,
        resource_id=resource_id,
        filters=filters,
    )
    result["result"] = ret["result"]
    result["comment"] = ret["comment"]
    if result["result"] and ret["ret"]:
        result["old_state"] = ret["ret"]
        # Populate both "old_state" and "new_state" with the same data
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


def convert_raw_db_parameter_group_to_present(
    hub,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
    parameters: List[Dict[str, Any]] = None,
    tags: Dict[str, Any] = None,
) -> Dict[str, Any]:
    """
    Convert AWS returned data structure to correct idem db_parameter_group present state

     Args:
        hub: required for functions in hub
        raw_resource: The aws response to convert
        idem_resource_name (Text, Optional): The idem of the idem resource
        Parameters (List[Dict[str, str]], Optional): The Parameter list of the resource. Defaults to None.
        tags (Dict, Optional): The tags of the resource. Defaults to None.

    Returns: Valid idem state for db_parameter_group of type Dict['string', Any]
    """
    resource_id = raw_resource.get("DBParameterGroupName")
    raw_resource["Tags"] = tags
    raw_resource["Parameters"] = parameters
    resource_parameters = OrderedDict(
        {
            "DBParameterGroupFamily": "db_parameter_group_family",
            "Description": "description",
            "Tags": "tags",
            "Parameters": "parameters",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw) is not None:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    return resource_translated



import copy
from typing import Any
from typing import Dict
from typing import List


async def update_tags(
    hub,
    ctx,
    resource_id,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
):
    """
    Update tags of AWS IAM OpenID Connect Provider

    Args:
        resource_id: The Amazon Resource Name (ARN) of the IAM OpenID Connect provider
        old_tags: dict of old tags
        new_tags: dict of new tags

    Returns:
        {"result": True|False, "comment": Tuple, "ret": Dict of updated tags}

    """
    tags_to_add = {}
    tags_to_remove = {}
    if new_tags is not None:
        tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )

    result = dict(comment=(), result=True, ret=None)
    if (not tags_to_remove) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result

    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.iam.untag_open_id_connect_provider(
                ctx,
                OpenIDConnectProviderArn=resource_id,
                TagKeys=list(tags_to_remove.keys()),
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result

    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.iam.tag_open_id_connect_provider(
                ctx,
                OpenIDConnectProviderArn=resource_id,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags=tags_to_add),
            )
            if not add_ret["result"]:
                result["comment"] = add_ret["comment"]
                result["result"] = False
                return result

    result["ret"] = new_tags
    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    else:
        result["comment"] = hub.tool.aws.comment_utils.update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    return result


async def update_thumbprints(
    hub,
    ctx,
    resource_id,
    old_thumbprints: List[str],
    new_thumbprints: List[str],
):
    """
    Update thumbprints of AWS IAM OpenID Connect Provider

    Args:
        resource_id: The Amazon Resource Name (ARN) of the IAM OpenID Connect provider
        old_thumbprints: list of old thumbprints
        new_thumbprints: list of new thumbprints

    Returns:
        {"result": True|False, "comment": Tuple, "ret": Dict}
    """
    result = dict(comment=(), result=True, ret=None)
    if not ctx.get("test", False):
        update_ret = (
            await hub.exec.boto3.client.iam.update_open_id_connect_provider_thumbprint(
                ctx=ctx,
                OpenIDConnectProviderArn=resource_id,
                ThumbprintList=new_thumbprints,
            )
        )
        if not update_ret["result"]:
            result["comment"] = update_ret["comment"]
            result["result"] = False
            return result

    result["ret"] = {"thumbprint_list": new_thumbprints}
    if ctx.get("test", False):
        result["comment"] = (f"Would update thumbprints: {new_thumbprints}",)
    else:
        result["comment"] = (f"Updated thumbprints: {new_thumbprints}",)
    return result


async def update_client_ids(
    hub,
    ctx,
    resource_id,
    old_client_ids: List[str],
    new_client_ids: List[str],
):
    """
    Update clientIDs of AWS IAM OpenID Connect Provider

    Args:
        resource_id: The Amazon Resource Name (ARN) of the IAM OpenID Connect provider
        old_client_ids: list of old clientIDs
        new_client_ids: list of new clientIDs

    Returns:
        {"result": True|False, "comment": Tuple, "ret": Dict}

    """
    old_clients_list = copy.deepcopy(old_client_ids)
    clients_result = copy.deepcopy(old_clients_list)

    clients_to_add = list(set(new_client_ids).difference(old_client_ids))
    clients_to_remove = list(set(old_client_ids).difference(new_client_ids))

    result = dict(comment=(), result=True, ret=None)
    if (not clients_to_remove) and (not clients_to_add):
        return result

    if clients_to_remove:
        if not ctx.get("test", False):
            for client in clients_to_remove:
                delete_ret = await hub.exec.boto3.client.iam.remove_client_id_from_open_id_connect_provider(
                    ctx, OpenIDConnectProviderArn=resource_id, ClientID=client
                )
                if not delete_ret["result"]:
                    result["comment"] = delete_ret["comment"]
                    result["result"] = False
                    return result
        [clients_result.remove(key) for key in clients_to_remove]

    if clients_to_add:
        if not ctx.get("test", False):
            for client in clients_to_add:
                add_ret = await hub.exec.boto3.client.iam.add_client_id_to_open_id_connect_provider(
                    ctx, OpenIDConnectProviderArn=resource_id, ClientID=client
                )
                if not add_ret["result"]:
                    result["comment"] = add_ret["comment"]
                    result["result"] = False
                    return result

    result["ret"] = {"client_id_list": clients_result + clients_to_add}
    if ctx.get("test", False):
        result["comment"] = (
            f"Would update clientIDs: Add ({clients_to_add}) Remove ({clients_to_remove})",
        )
    else:
        result["comment"] = (
            f"Updated clientIDs: Add ({clients_to_add}) Remove ({clients_to_remove})",
        )
    return result



"""
Util functions to execute update on AWS DynamoDB Table.
"""
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


async def compare_inputs_and_update_dynamodb_table(
    hub,
    ctx,
    plan_state: Dict[str, Any],
    attribute_definitions: List,
    name: str,
    billing_mode: str = None,
    provisioned_throughput: dict = None,
    global_secondary_indexes: List = None,
    sse_specification: dict = None,
    stream_specification: dict = None,
    replica_updates: List = None,
    table_class: str = None,
    timeout: Dict = None,
):
    result = dict(comment=(), result=True, ret=None)
    if not ctx.get("test", False):
        update_ret = await hub.exec.boto3.client.dynamodb.update_table(
            ctx=ctx,
            AttributeDefinitions=attribute_definitions,
            TableName=name,
            BillingMode=billing_mode,
            ProvisionedThroughput=provisioned_throughput,
            GlobalSecondaryIndexes=global_secondary_indexes,
            SSESpecification=sse_specification,
            StreamSpecification=stream_specification,
            ReplicaUpdates=replica_updates,
            TableClass=table_class,
        )
        result["result"] = update_ret["result"]
        result["comment"] = (
            update_ret["comment"]
            if result["result"]
            else (f"Updated aws.dynamodb.table {name}.",)
        )
        if update_ret["result"]:
            waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=15,
                default_max_attempts=40,
                timeout_config=timeout.get("update") if timeout else None,
            )
            hub.log.debug(f"Waiting on updating aws.dynamodb.table '{name}'")
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "dynamodb",
                    "table_exists",
                    TableName=name,
                    WaiterConfig=waiter_config,
                )
            except Exception as e:
                result["comment"] = result["comment"] + (str(e),)
                result["result"] = False
    else:
        update_params = OrderedDict(
            {
                "billing_mode": billing_mode,
                "attribute_definitions": attribute_definitions,
                "name": name,
                "table_class": table_class,
                "provisioned_throughput": provisioned_throughput,
                "global_secondary_indexes": global_secondary_indexes,
                "sse_specification": sse_specification,
                "stream_specification": stream_specification,
                "replica_updates": replica_updates,
            }
        )
        for key, value in update_params.items():
            if value is None:
                if key not in plan_state:
                    plan_state[key] = None
            else:
                plan_state[key] = value
    return result



import boto3.session


def __init__(hub):
    # Create a single session for everything to be run from
    hub.tool.boto3.SESSION = boto3.session.Session()


def get(hub, botocore_session=None) -> boto3.session.Session:
    """
    Get the current boto3 session.
    """
    if botocore_session is None:
        # Reset the session if need be for thread safety
        hub.tool.boto3.SESSION = boto3.session.Session()
    return hub.tool.boto3.SESSION



import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    user_name: str,
    resource_id: str = None,
    status: str = "Active",
    pgp_key: str = None,
    secret_access_key: str = None,
) -> Dict[str, Any]:
    r"""
    Ensures an AWS key has the assigned status.

    This will create a new access key for a user if no access_key_id is passed for this state, either via
    access_key_id or resource_id.

    If a new access key is created, the secret access key will also be returned. This can optionally be encrypted
    with a base 64 encoded PGP public key.

    Args:
        name(Text): An Idem name describing the resource.
        user_name(Text): AWS IAM user name that the key belongs to.
        resource_id(Text, optional): AWS IAM access key ID.
        status(Text, optional): "Active" or "Inactive"; Active keys are valid for API calls. Defaults to "Active".
        pgp_key(Text, optional): A base 64 encode PGP public key, used to encrypt the secret access key if a new key is created.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls
            name_describing_key:
              aws.iam.access_key.present:
                - user_name: aws_user
                - status: Active
    """
    # secret_access_key is used internally to make sure the secret_access_key isn't lost after the 1st present run.
    # This parameter should not be inputted on sls and hence it is not added to the function description.
    result = {
        "name": name,
        "old_state": None,
        "new_state": None,
        "comment": (),
        "result": True,
    }

    # ----- look up key
    if resource_id:
        list_ret = await hub.exec.aws.iam.access_key.list(
            ctx, access_key_id=resource_id, user_name=user_name
        )
        if not list_ret["result"]:
            result["result"] = False
            result["comment"] += (f"Error listing access keys",) + list_ret["comment"]
            return result
        elif len(list_ret["ret"]) == 0:
            # Return error if the specified access key id doesn't exist on AWS.
            message = (
                f"The specified access_key_id {resource_id} in aws.iam.key {name} does not exist, and keys "
                "cannot be recreated by id."
            )
            result["comment"] += (message,) + list_ret["comment"]
            result["result"] = False
            return result

        before = list_ret["ret"][0]
        if secret_access_key:
            before["secret_access_key"] = secret_access_key
        result[
            "old_state"
        ] = hub.tool.aws.iam.conversion_utils.convert_access_key_to_present(
            before, name
        )
        if before["status"] == status:
            result["comment"] += (f"No changes necessary for aws.iam.key {name}",)
            result["new_state"] = copy.deepcopy(result["old_state"])
            return result
        else:
            # only update is necessary
            if not ctx.get("test", False):
                update_ret = await hub.exec.aws.iam.access_key.update(
                    ctx, user_name=user_name, access_key_id=resource_id, status=status
                )
                if not update_ret["result"]:
                    result["comment"] += update_ret["comment"]
                    result["result"] = False
                    return result
                result["comment"] += (f"Updated aws.iam.key {name} to {status}",)
            else:
                result["comment"] += (f"Would update aws.iam.key {name} to {status}",)
            result["new_state"] = copy.deepcopy(result["old_state"])
            result["new_state"]["status"] = status
    else:
        if ctx.get("test"):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": "GENERATED_KEY",
                    "user_name": user_name,
                    "secret_access_key": "SECRET_ACCESS_KEY",
                    "status": status if status else "Active",
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.iam.access_key", name=name
            )
            return result
        create_ret = await hub.exec.aws.iam.access_key.create(ctx, user_name, pgp_key)
        if not create_ret["result"]:
            result["result"] = False
            result["comment"] += create_ret["comment"]
            return result
        access_key_id = create_ret["ret"]["access_key_id"]
        result[
            "new_state"
        ] = hub.tool.aws.iam.conversion_utils.convert_access_key_to_present(
            create_ret["ret"], name
        )
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.iam.access_key", name=name
        )
        # Create can't set status. So update the status if the status should be "Inactive"
        if status == "Inactive":
            update_ret = await hub.exec.aws.iam.access_key.update(
                ctx, user_name=user_name, access_key_id=access_key_id, status=status
            )
            if not update_ret["result"]:
                result["comment"] += update_ret["comment"]
                result["result"] = False
            result["new_state"]["status"] = "Inactive"
    return result


async def absent(
    hub,
    ctx,
    name: str,
    user_name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    Ensure the specified access key does not exist.
    Either resource_id or both user_name and access_key_id must be passed in.

    Args:
        name(Text): An Idem name describing the resource.
        user_name(Text): AWS IAM user name that the key belongs to.
        resource_id(Text, optional): AWS IAM access key ID.
    Returns:
        Dict[str, Any]
    Examples:
        .. code-block:: sls
            name_describing_key:
              aws.iam.access_key.absent:
                - user_name: value
                - resource_id: value
    """
    result = {
        "name": name,
        "old_state": None,
        "new_state": None,
        "comment": (),
        "result": True,
    }

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.iam.access_key", name=name
        )
        return result

    # ------ look up key
    list_ret = await hub.exec.aws.iam.access_key.list(
        ctx, access_key_id=resource_id, user_name=user_name
    )
    result["comment"] = list_ret["comment"]
    if not list_ret["result"]:
        result["result"] = False
        return result
    if len(list_ret["ret"]) == 0:
        result["comment"] += hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.iam.access_key", name=name
        )
        return result

    access_key = list_ret["ret"][0]
    result[
        "old_state"
    ] = hub.tool.aws.iam.conversion_utils.convert_access_key_to_present(
        access_key, name
    )

    # ------ delete key
    # delete logic is handled inside exec's delete, including ctx["test"]
    if ctx.get("test"):
        result["comment"] += hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.iam.access_key", name=name
        )
        return result
    delete_ret = await hub.exec.aws.iam.access_key.delete(
        ctx, access_key_id=resource_id, user_name=user_name
    )
    result["result"] = delete_ret["result"]
    if not result["result"]:
        result["comment"] += delete_ret["comment"]
    else:
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.iam.access_key", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describe access keys and their current status in a way that can be managed via the "present" function.

    We describe all access keys for all users the logged in user can list.

    Returns:
        Dict[str, Any]
    Examples:
        .. code-block:: bash
            $ idem describe aws.iam.access_key
    """
    users_list = await hub.exec.aws.iam.user.list(ctx)
    if not users_list["result"]:
        hub.log.debug(f"Could not describe keys: {users_list['comment']}")
        return {}
    users = [u["user_name"] for u in users_list["ret"]]

    result = {}

    for user in users:
        result.update(await _describe_one_user(hub, ctx, user))
    # The above is pretty slow, if/when we add random exponential backoff on
    #  failures, switch to this code:
    # multi_result = await asyncio.gather(
    #     *[_describe_one_user(hub, ctx, u["user_name"]) for u in users]
    # )
    # for r in multi_result:
    #     result.update(r)

    return result


async def _describe_one_user(hub, ctx, user_name: str) -> Dict[str, Dict[str, Any]]:
    keys = await hub.exec.aws.iam.access_key.list(ctx, user_name=user_name)
    if not keys["result"]:
        hub.log.debug(
            f"Could not list aws.iam.key access keys for user {user_name}: {keys['comment']}"
        )
        return {}

    result = {}

    for access_key in keys["ret"]:
        # iterate every loop to give the user another indication if keys are skipped
        if "access_key_id" not in access_key or "status" not in access_key:
            # All values are "optionally" returned from AWS but managing a key
            #  without the id and status would be impossible.
            hub.log.debug(f"Can not describe an aws key without an id and status")
            continue

        idem_name = f"{access_key['user_name']}-{access_key['access_key_id']}"

        state = hub.tool.aws.iam.conversion_utils.convert_access_key_to_present(
            access_key, idem_name
        )
        resource_key = f"iam-access-key-{idem_name}"
        result[resource_key] = {
            "aws.iam.access_key.present": [{k: v} for k, v in state.items()]
        }

    return result



from typing import Any
from typing import Dict


"""
Util functions for AWS S3 bucket versioning resources.
"""


def convert_raw_bucket_versioning_to_present(
    hub, bucket: str, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    r"""
    Convert an AWS S3 bucket versioning resource to a common idem present state.

    Args:
        hub: required for functions in hub.
        bucket(string): The S3 bucket name in Amazon Web Services.
        raw_resource(Dict[str, Any]): The AWS response to convert.
        idem_resource_name(string, optional): An Idem name of the resource.

    Returns:
        Dict[str, Any]: Common idem present state.
    """

    raw_resource.pop("ResponseMetadata", None)

    resource_translated = {
        "name": idem_resource_name if idem_resource_name else f"{bucket}-versioning",
        "resource_id": bucket,
        "bucket": bucket,
        "mfa_delete": raw_resource.get("MFADelete", "Disabled"),
        "status": raw_resource.get("Status", "Suspended"),
    }

    return resource_translated



from typing import Dict

from dict_tools import differ


async def update_topic_attributes(
    hub,
    ctx,
    resource_arn: str,
    old_attributes: Dict[str, str],
    new_attributes: Dict[str, str],
):
    """
    Given old and new attributes of SNS topic the function updates the attributes.

    Args:
        hub:
        ctx:
        resource_arn(Text): AWS resource_arn of SNS topic
        old_attributes(Dict): old attribute of SNS topic
        new_attributes(Dict): new attribute of SNS topic

    """
    result = dict(comment=(), result=True, ret=None)
    attributes_diff = differ.deep_diff(old_attributes, new_attributes)
    attributes_to_update = attributes_diff.get("new")

    if ctx.get("test", False):
        result["ret"] = {"updated_attributes": attributes_to_update}
        result["comment"] = (f"Would Update attributes {attributes_to_update.keys()}",)
        return result

    else:
        if attributes_to_update:
            for key, value in attributes_to_update.items():
                ret = await hub.exec.boto3.client.sns.set_topic_attributes(
                    ctx, TopicArn=resource_arn, AttributeName=key, AttributeValue=value
                )
                if not ret["result"]:
                    result["comment"] = ret["comment"]
                    result["result"] = False
                    return result

            result["ret"] = {"updated_attributes": attributes_to_update}
            result["comment"] = (f"Updated attributes {attributes_to_update.keys()}",)
    return result


async def update_subscription_attributes(
    hub,
    ctx,
    resource_arn: str,
    old_attributes: Dict[str, str],
    new_attributes: Dict[str, str],
):
    """
    Given old and new attributes of SNS topic_subscription the function updates the attributes.

    Args:
        hub:
        ctx:
        resource_arn(Text): AWS resource_arn of SNS topic_subscription
        old_attributes(Dict): old attribute of SNS topic_subscription
        new_attributes(Dict): new attribute of SNS topic_subscription

    """
    result = dict(comment=(), result=True, ret=None)
    attributes_diff = differ.deep_diff(old_attributes, new_attributes)
    attributes_to_update = attributes_diff.get("new")

    if ctx.get("test", False):
        result["ret"] = {"updated_attributes": attributes_to_update}
        result["comment"] = (f"Would Update attributes {attributes_to_update.keys()}",)
        return result

    else:
        if attributes_to_update:
            for key, value in attributes_to_update.items():
                ret = await hub.exec.boto3.client.sns.set_subscription_attributes(
                    ctx,
                    SubscriptionArn=resource_arn,
                    AttributeName=key,
                    AttributeValue=value,
                )
                if not ret["result"]:
                    result["comment"] = ret["comment"]
                    result["result"] = False
                    return result

            result["ret"] = {"updated_attributes": attributes_to_update}
            result["comment"] = (f"Updated attributes {attributes_to_update.keys()}",)
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.lambda.create_function
hub.exec.boto3.client.lambda.delete_function
hub.exec.boto3.client.lambda.get_function
hub.exec.boto3.client.lambda.update_function_code

"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


async def present(
    hub,
    ctx,
    name: str,
    role: str,
    code: make_dataclass(
        "FunctionCode",
        [
            ("ZipFile", bytes, field(default=None)),
            ("S3Bucket", str, field(default=None)),
            ("S3Key", str, field(default=None)),
            ("S3ObjectVersion", str, field(default=None)),
            ("ImageUri", str, field(default=None)),
        ],
    ),
    resource_id: str = None,
    runtime: str = None,
    handler: str = None,
    description: str = None,
    function_timeout: int = None,
    memory_size: int = None,
    publish: bool = None,
    vpc_config: make_dataclass(
        "VpcConfig",
        [
            ("SubnetIds", List[str], field(default=None)),
            ("SecurityGroupIds", List[str], field(default=None)),
        ],
    ) = None,
    package_type: str = None,
    dead_letter_config: make_dataclass(
        "DeadLetterConfig", [("TargetArn", str, field(default=None))]
    ) = None,
    environment: make_dataclass(
        "Environment", [("Variables", Dict[str, str], field(default=None))]
    ) = None,
    kms_key_arn: str = None,
    tracing_config: make_dataclass(
        "TracingConfig", [("Mode", str, field(default=None))]
    ) = None,
    tags: Dict[str, str] = None,
    layers: List[str] = None,
    file_system_configs: List[
        make_dataclass("FileSystemConfig", [("Arn", str), ("LocalMountPath", str)])
    ] = None,
    image_config: make_dataclass(
        "ImageConfig",
        [
            ("EntryPoint", List[str], field(default=None)),
            ("Command", List[str], field(default=None)),
            ("WorkingDirectory", str, field(default=None)),
        ],
    ) = None,
    code_signing_config_arn: str = None,
    architectures: List[str] = None,
    qualifier: str = None,
    maximum_retry_attempts: str = None,
    maximum_event_age_in_seconds: str = None,
    destination_config: str = None,
    revision_id: str = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    r"""

    Creates a Lambda function. To create a function, you need a deployment package and an execution role . The
    deployment package is a .zip file archive or container image that contains your function code. The execution role
    grants the function permission to use Amazon Web Services, such as Amazon CloudWatch Logs for log
    streaming and X-Ray for request tracing.

    When you create a function, Lambda provisions an instance of the function and its supporting resources. If your
    function connects to a VPC, this process can take a minute or so. During this time, you can't invoke or modify
    the function. The State , StateReason , and StateReasonCode fields in the response from GetFunctionConfiguration
    indicate when the function is ready to invoke. For more information, see Function States .

    A function has an unpublished version, and can have published versions and aliases. The unpublished version
    changes when you update your function's code and configuration. A published version is a snapshot of your
    function code and configuration that can't be changed. An alias is a named resource that maps to a version,
    and can be changed to map to a different version. Use the Publish parameter to create version 1 of your function
    from its initial configuration.

    Args:
        resource_id(Text): The name/ ARN/ partial ARN of the AWS Lambda function, version, or alias.
        name(Text): The name of the Lambda function.
            Name formats
                Function name - my-function .
                Function ARN - arn:aws:lambda:us-west-2:123456789012:function:my-function .
                Partial ARN - 123456789012:function:my-function .
            The length constraint applies only to the full ARN. If you specify only the function name, it is limited
                to 64 characters in length.
        role(Text): The Amazon Resource Name (ARN) of the function's execution role.
        code(Dict[str, Any]): The code for the function.
            * ZipFile (ByteString, optional): The base64-encoded contents of the deployment package. Amazon Web Services SDK and Amazon Web
                Services CLI clients handle the encoding for you.
            * S3Bucket (str, optional): An Amazon S3 bucket in the same Amazon Web Services Region as your function. The bucket can be
                in a different Amazon Web Services account.
            * S3Key (str, optional): The Amazon S3 key of the deployment package.
            * S3ObjectVersion (str, optional): For versioned objects, the version of the deployment package object to use.
            * ImageUri (str, optional): URI of a container image in the Amazon ECR registry.
        runtime(Text, optional): The identifier of the function's runtime . Runtime is required if the deployment package
            is a .zip file archive.
        revision_id(Text, optional): Only update the function if the revision ID matches the ID that's specified.
                Use this option to avoid modifying a function that has changed since you last read it.
        destination_config(dict, optional): A destination for events after they have been sent to a function for
                processing.
        maximum_event_age_in_seconds(int, optional): The maximum age of a request that Lambda sends to a function for
                processing.
        maximum_retry_attempts (int, optional): The maximum number of times to retry when the function returns an error.
        qualifier(Text): The version of the Lambda function.
        architectures(list, optional): The instruction set architecture that the function supports. Enter a string array
                with one of the valid values (arm64 or x86_64). The default value is x86_64 .
        code_signing_config_arn(Text, optional): To enable code signing for this function, specify the ARN of a
                code-signing configuration. A code-signing configuration includes a set of signing profiles, which
                define the trusted publishers for this function.
        image_config(Dict[str, Any], optional): Container image configuration values that override the values in the container image Dockerfile. Defaults to None.
            * EntryPoint (List[str], optional): Specifies the entry point to their application, which is typically the location of the runtime
                executable.
            * Command (List[str], optional): Specifies parameters that you want to pass in with ENTRYPOINT.
            * WorkingDirectory (str, optional): Specifies the working directory.
        file_system_configs(List[Dict[str, Any]], optional): Connection settings for an Amazon EFS file system. Defaults to None.
            * Arn (str): The Amazon Resource Name (ARN) of the Amazon EFS access point that provides access to the file
                system.
            * LocalMountPath (str): The path where the function can access the file system, starting with /mnt/.
        layers(list, optional): A list of function layers to add to the function's execution environment. Specify each
                layer by its ARN, including the version.
        tags(Dict[str, str], optional): A list of tags to apply to the function. Defaults to None.
        tracing_config(Dict[str, Any], optional): Set Mode to Active to sample and trace a subset of incoming requests with X-Ray. Defaults to None.
            * Mode (str, optional): The tracing mode.
        kms_key_arn(Text, optional): The ARN of the Amazon Web Services Key Management Service (KMS) key that's used to
                encrypt your function's environment variables. If it's not provided, Lambda uses a default service key.
        environment(Dict[str, Any], optional): Environment variables that are accessible from function code during execution. Defaults to None.
            * Variables (Dict[str, str], optional): Environment variable key-value pairs. For more information, see Using Lambda environment
                variables.
        dead_letter_config(Dict[str, Any], optional): A dead letter queue configuration that specifies the queue or topic where Lambda sends
            asynchronous events when they fail processing. For more information, see Dead Letter Queues. Defaults to None.
            * TargetArn (str, optional): The Amazon Resource Name (ARN) of an Amazon SQS queue or Amazon SNS topic.
        package_type(Text, optional): The type of deployment package. Set to Image for container image and set Zip for
                ZIP archive.
        vpc_config(Dict[str, Any], optional): For network connectivity to Amazon Web Services resources in a VPC, specify a list of security
            groups and subnets in the VPC. When you connect a function to a VPC, it can only access
            resources and the internet through that VPC. For more information, see VPC Settings. Defaults to None.
            * SubnetIds (List[str], optional): A list of VPC subnet IDs.
            * SecurityGroupIds (List[str], optional): A list of VPC security groups IDs.
        publish(bool, optional): Set to true to publish the first version of the function during creation.
        memory_size(int, optional): The amount of memory available to the function at runtime. Increasing the function
                memory also increases its CPU allocation. The default value is 128 MB. The value can be any multiple of
                1 MB.
        function_timeout(int, optional): The amount of time (in seconds) that Lambda allows a function to run before
                stopping it. The default is 3 seconds. The maximum allowed value is 900 seconds.
        description(Text, optional): A description of the function.
        handler(Text, optional): The name of the method within your code that Lambda calls to execute your function.
                Handler is required if the deployment package is a .zip file archive. The format includes the file name.
                It can also include namespaces and other qualifiers, depending on the runtime.
        timeout(Dict, optional): Timeout configuration for create/update/deletion of AWS IAM Policy.
            * create (Dict): Timeout configuration for creating AWS IAM Policy
                * delay (int, Optional): The amount of time in seconds to wait between attempts.
                * max_attempts (int, Optional): Customized timeout configuration containing delay and max attempts.
            * update(Dict, optional): Timeout configuration for updating AWS IAM Policy
                * delay (int, Optional): The amount of time in seconds to wait between attempts.
                * max_attempts: (int, Optional) Customized timeout configuration containing delay and max attempts.
        hub:
        ctx:

    Request Syntax:
        [lambda_function-id]:
          aws.lambda.function.present:
          - name: 'string'
          - runtime: 'string'
          - role: 'string'
          - code: 'Dict'
          - handler: 'string'
          - description: 'string'
          - function_timeout: 'int'
          - memory_size: 'int'
          - publish: 'bool'
          - vpc_config: 'Dict'
          - package_type: 'string'
          - dead_letter_config: 'string'
          - environment: 'Dict'
          - kms_key_arn: 'string'
          - tracing_config: 'Dict'
          - layers: 'list'
          - file_system_configs: 'list'
          - image_config: 'Dict'
          - code_signing_config_arn: 'string'
          - architectures: 'list'
          - version: 'string'
          - tags:
            - Key: 'string'
              Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        . code-block:: sls

            test_dyanmodb_table:
              aws.lambda.function.present:
                - name: test_dyanmodb_table
                - resource_id: test_dyanmodb_table
                - run_time: nodejs12.x
                - role: arn:aws:iam::123456789012:role/lambda-role
                - code:{
                           'S3Bucket': 'my-bucket-1xpuxmplzrlbh',
                            'S3Key': 'function.zip',
                        },
                - tags:
                  - Key: DEPARTMENT
                    Value: Assets
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    are_tags_updated = False
    is_function_updated = False
    existing_tags = {}
    plan_state = {}
    function_arn = None
    before = await hub.exec.boto3.client["lambda"].get_function(
        ctx=ctx, FunctionName=resource_id, Qualifier=qualifier
    )
    if before["result"] and before["ret"].get("Configuration"):
        function = before["ret"].get("Configuration")
        code_config = before["ret"].get("Code")
        function_arn = function.get("FunctionArn")
        tags_return = await hub.exec.boto3.client["lambda"].list_tags(
            ctx, Resource=function_arn
        )
        result["result"] = tags_return.get("result")
        if result["result"]:
            existing_tags = tags_return.get("ret").get("Tags")
        else:
            result["comment"] = result["comment"] + tags_return["comment"]
            return result
        resource_converted = await hub.tool.aws.lambda_aws.conversion_utils.convert_raw_lambda_function_to_present(
            ctx,
            raw_resource=function,
            idem_resource_name=name,
            tags=existing_tags,
        )
        result["result"] = result["result"] and resource_converted["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + resource_converted["comment"]
        result["old_state"] = resource_converted["ret"]
        plan_state = copy.deepcopy(result["old_state"])

        update = {
            "FunctionName": name,
            "ZipFile": code.get("ZipFile"),
            "S3Bucket": code.get("S3Bucket"),
            "S3Key": code.get("S3Key"),
            "S3ObjectVersion": code.get("S3ObjectVersion"),
            "ImageUri": code.get("ImageUri"),
            "Publish": publish,
            "RevisionId": revision_id,
            "Architectures": architectures,
            "Role": role,
            "Handler": handler,
            "Description": description,
            "Timeout": function_timeout,
            "MemorySize": memory_size,
            "VpcConfig": vpc_config,
            "Environment": environment,
            "Runtime": runtime,
            "DeadLetterConfig": dead_letter_config,
            "KMSKeyArn": kms_key_arn,
            "TracingConfig": tracing_config,
            "Layers": layers,
            "Code": code,
            "Tags": tags,
            "FileSystemConfigs": file_system_configs,
            "ImageConfig": image_config,
            "Qualifier": qualifier,
            "MaximumRetryAttempts": maximum_retry_attempts,
            "MaximumEventAgeInSeconds": maximum_event_age_in_seconds,
            "DestinationConfig": destination_config,
            "CodeSigningConfigArn": code_signing_config_arn,
            "PackageType": package_type,
        }
        update_return = await hub.exec.aws.lambda_aws.function.compare_inputs_and_update_lambda_function(
            ctx=ctx,
            name=name,
            function=function,
            update=update,
            code=code_config,
            timeout=timeout,
            plan_state=plan_state,
            resource_id=resource_id,
        )
        result["result"] = result["result"] and update_return["result"]
        result["comment"] = result["comment"] + update_return["comment"]
        if result["result"]:
            is_function_updated = True

        if tags is not None:
            # Update tags
            update_tags_result = await hub.exec.aws.lambda_aws.tag.update_tags(
                ctx=ctx,
                resource_arn=function_arn,
                old_tags=existing_tags,
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_tags_result["comment"]
            result["result"] = result["result"] and update_tags_result["result"]

            if not result["result"]:
                return result
            are_tags_updated = result["result"]

            if ctx.get("test", False) and (update_tags_result["ret"] is not None):
                plan_state["tags"] = update_tags_result["ret"].get("tags")
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "role": role,
                    "code": code,
                    "resource_id": name,
                    "runtime": runtime,
                    "handler": handler,
                    "description": description,
                    "function_timeout": function_timeout,
                    "memory_size": memory_size,
                    "publish": publish,
                    "vpc_config": vpc_config,
                    "package_type": package_type,
                    "dead_letter_config": dead_letter_config,
                    "environment": environment,
                    "kms_key_arn": kms_key_arn,
                    "tracing_config": tracing_config,
                    "tags": tags,
                    "layers": layers,
                    "file_system_configs": file_system_configs,
                    "image_config": image_config,
                    "code_signing_config_arn": code_signing_config_arn,
                    "architectures": architectures,
                    "qualifier": qualifier,
                    "maximum_retry_attempts": maximum_retry_attempts,
                    "maximum_event_age_in_seconds": maximum_event_age_in_seconds,
                    "destination_config": destination_config,
                    "revision_id": revision_id,
                },
            )
            result["comment"] = (f"Would create aws.lambda.function {name}",)
            return result
        try:
            ret = await hub.exec.boto3.client["lambda"].create_function(
                ctx,
                FunctionName=name,
                Runtime=runtime,
                Role=role,
                Handler=handler,
                Code=code,
                Description=description,
                Timeout=function_timeout,
                MemorySize=memory_size,
                Publish=publish,
                VpcConfig=vpc_config,
                PackageType=package_type,
                DeadLetterConfig=dead_letter_config,
                Environment=environment,
                KMSKeyArn=kms_key_arn,
                TracingConfig=tracing_config,
                Tags=tags,
                Layers=layers,
                FileSystemConfigs=file_system_configs,
                ImageConfig=image_config,
                CodeSigningConfigArn=code_signing_config_arn,
                Architectures=architectures,
            )
            result["result"] = ret["result"]
            if not ret["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                return result

            waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=1,
                default_max_attempts=20,
                timeout_config=timeout.get("create") if timeout else None,
            )
            hub.log.debug(f"Waiting on creating aws.lambda.function '{name}'")
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "lambda",
                    "function_active",
                    FunctionName=ret["ret"]["FunctionArn"],
                    WaiterConfig=waiter_config,
                )
            except Exception as e:
                result["comment"] = result["comment"] + (str(e),)
                result["result"] = False
            resource_id = ret["ret"]["FunctionName"]
            function_arn = ret["ret"]["FunctionArn"]
            existing_tags = tags
            ret["comment"] = result["comment"] + (
                f"Created aws.lambda.function '{resource_id}'",
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif not (before and before["result"]) or is_function_updated:
        after = await hub.exec.boto3.client["lambda"].get_function(
            ctx, FunctionName=resource_id, Qualifier=qualifier
        )
        final_updated_tags = existing_tags
        if are_tags_updated:
            tags = await hub.exec.boto3.client["lambda"].list_tags(
                ctx, Resource=function_arn
            )
            if tags.get("result") and tags.get("ret"):
                final_updated_tags = tags.get("ret").get("Tags")
        resource_converted = await hub.tool.aws.lambda_aws.conversion_utils.convert_raw_lambda_function_to_present(
            ctx=ctx,
            raw_resource=after["ret"].get("Configuration"),
            idem_resource_name=name,
            tags=final_updated_tags,
        )
        result["result"] = result["result"] and resource_converted["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + resource_converted["comment"]
        result["new_state"] = resource_converted["ret"]
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str,
    qualifier: str = None,
) -> Dict[str, Any]:
    r"""

    Deletes a Lambda function. To delete a specific function version, use the Qualifier parameter.
    Otherwise, all versions and aliases are deleted.

    To delete Lambda event source mappings that invoke a function, use DeleteEventSourceMapping . For Amazon Web
    Service services and resources that invoke your function directly, delete the trigger in the service where you
    originally configured it.

    Args:
        resource_id(Text): The name/ ARN/ partial ARN of the AWS Lambda function, version, or alias.
        qualifier(Text, optional): Specify a version or alias to get details about a published version of the function.
        name(Text): The name of the Lambda function, version, or alias.
            Name formats
                Function name - my-function (name-only), my-function:v1 (with alias).
                Function ARN - arn:aws:lambda:us-west-2:123456789012:function:my-function .
                Partial ARN - 123456789012:function:my-function .
        hub:
        ctx:

    Request Syntax:
        [lambda_function-id]:
          aws.lambda.function.absent:
          - name: 'string'
          - qualifier: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        . code-block:: sls

            resource_is_absent:
                aws.lambda.function.absent:
                    - name: resource_is_absent
                    - qualifier: 1
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.boto3.client["lambda"].get_function(
        ctx, FunctionName=resource_id, Qualifier=qualifier
    )
    if not before["result"]:
        result["comment"] = (f"aws.lambda.function '{name}' already absent",)
    else:
        tags = await hub.exec.boto3.client["lambda"].list_tags(
            ctx, Resource=before["ret"].get("Configuration").get("FunctionArn")
        )
        if tags.get("result"):
            existing_tags = tags.get("ret").get("Tags")
        else:
            result["comment"] = result["comment"] + tags["comment"]
            result["result"] = False
            return result

        resource_converted = await hub.tool.aws.lambda_aws.conversion_utils.convert_raw_lambda_function_to_present(
            ctx,
            raw_resource=before["ret"].get("Configuration"),
            idem_resource_name=resource_id,
            tags=existing_tags,
        )
        result["result"] = result["result"] and resource_converted["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + resource_converted["comment"]
        result["old_state"] = resource_converted["ret"]
        if ctx.get("test", False):
            result["comment"] = f"Would delete aws.lambda.function {resource_id}"
            return result
        else:
            try:
                ret = await hub.exec.boto3.client["lambda"].delete_function(
                    ctx, FunctionName=resource_id, Qualifier=qualifier
                )
                result["result"] = ret["result"]
                if not result["result"]:
                    result["comment"] = ret["comment"]
                    return result
                result["comment"] = result["comment"] + (
                    f"Deleted aws.lambda.function '{name}'",
                )
                after = await hub.exec.boto3.client["lambda"].get_function(
                    ctx, FunctionName=resource_id, Qualifier=qualifier
                )
                if after["result"] and after["ret"]:
                    resource_converted = await hub.tool.aws.lambda_aws.conversion_utils.convert_raw_lambda_function_to_present(
                        ctx,
                        raw_resource=after["ret"].get("Configuration"),
                        idem_resource_name=resource_id,
                        tags=existing_tags,
                    )
                    result["result"] = result["result"] and resource_converted["result"]
                    if not result["result"]:
                        result["comment"] = (
                            result["comment"] + resource_converted["comment"]
                        )
                    result["new_state"] = resource_converted["ret"]
            except hub.tool.boto3.exception.ClientError as e:
                result["comment"] = result["comment"] + (
                    f"{e.__class__.__name__}: {e}",
                )
                result["result"] = False
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""

    Returns a list of Lambda functions, with the version-specific configuration of each. Lambda returns up to 50
    functions per call. Set FunctionVersion to ALL to include all published versions of each function in addition to
    the unpublished version.

    Args:
        hub:
        ctx:

    Request Syntax:
        [lambda_function-id]:
          aws.lambda.function.describe:
          - name: 'string'
          - qualifier: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        . code-block:: sls

            test_dyanmodb_table:
                aws.lambda.function.describe:
                    - name: test_dyanmodb_table
                    - qualifier: 1
    """

    result = {}

    ret = await hub.exec.boto3.client["lambda"].list_functions(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe function {ret['comment']}")
        return result
    if ret["ret"] and ret["ret"].get("Functions"):
        for function in ret["ret"].get("Functions"):
            tags = await hub.exec.boto3.client["lambda"].list_tags(
                ctx, Resource=function.get("FunctionArn")
            )
            resource_converted = await hub.tool.aws.lambda_aws.conversion_utils.convert_raw_lambda_function_to_present(
                ctx,
                raw_resource=function,
                idem_resource_name=function.get("FunctionName"),
                tags=tags.get("ret").get("Tags") if tags.get("result") else None,
            )
            if not resource_converted["result"]:
                hub.log.warning(
                    f"Could not describe aws.lambda.function '{function.get('FunctionName')}' "
                    f"with error {resource_converted['comment']}"
                )
            else:
                result[function.get("FunctionName")] = {
                    "aws.lambda.function.present": [
                        {parameter_key: parameter_value}
                        for parameter_key, parameter_value in resource_converted[
                            "ret"
                        ].items()
                    ]
                }
    return result



from typing import List


async def apply(
    hub, ctx, resource, *, old_value, new_value, comments: List[str]
) -> bool:
    """
    Modify an ec2 instance based on a single parameter in it's "present" state

    Args:
        hub:
        ctx: The ctx from a state module call
        resource: An ec2 instance resource object
        old_value: The previous value from the attributes of an existing instance
        new_value: The desired value from the ec2 instance present state parameters
        comments: A running list of comments abound the update process
    """
    # Can this even be modified? until we know throw an error if someone tries to modify this attribute
    return False



from typing import Any
from typing import Dict


def render_data_sources(hub, detector_describe: Dict[str, Any]) -> Dict[str, Any]:
    """
    The syntax for data_sources parameter is different in create and update scenarios. So to make it similar we are
    rendering the data  so that it can be used in update and describe functions.

    Returns:
        Dict

    """
    data_sources_new = {}
    kubernetes_source = {}
    if "DataSources" in detector_describe:
        if "Kubernetes" in detector_describe["DataSources"]:
            kubernetes_source = detector_describe["DataSources"].pop("Kubernetes")
            if kubernetes_source:
                data_sources_new["Kubernetes"] = {
                    "AuditLogs": {
                        "Enable": kubernetes_source["AuditLogs"]["Status"] == "ENABLED"
                    }
                }
        for key, val in detector_describe["DataSources"].items():
            data_sources_new[key] = {}
            data_sources_new[key] = {"Enable": val["Status"] == "ENABLED"}
        if kubernetes_source:
            # This is to preserve the input value. We don't have to have this, but it is safer.
            detector_describe["DataSources"]["Kubernetes"] = kubernetes_source
    return data_sources_new


def compare_dicts(hub, data_sources: Dict[str, Any], data_sources_new: Dict[str, Any]):
    """
    This functions helps in comparing two dicts.
    It compares each key value in both the dicts and return true or false based on the comparison

    Returns:
        {True|False}

    """

    for key, value in data_sources.items():
        if key in data_sources_new:
            if isinstance(data_sources[key], dict):
                if not compare_dicts(hub, data_sources[key], data_sources_new[key]):
                    return False
            elif value != data_sources_new[key]:
                return False
        else:
            return False
    return True



from collections import OrderedDict
from typing import Any
from typing import Dict

from dict_tools import differ


async def compare_inputs_and_update_lambda_function(
    hub,
    ctx,
    name: str,
    function: Dict[str, Any],
    update: Dict[str, Any],
    code: Dict[str, Any],
    resource_id: str,
    plan_state: Dict[str, Any],
    timeout: Dict = None,
):
    """
    Updates a Lambda function's code. If code signing is enabled for the function, the code package must be signed by
    a trusted publisher. The function's code is locked when you publish a version. You can't modify the code of a
    published version, only the unpublished version.

    Modify the version-specific settings of a Lambda function. When you update a function, Lambda provisions an
    instance of the function and its supporting resources. If your function connects to a VPC, this process can take
    a minute. During this time, you can't modify the function, but you can still invoke it. Updates the configuration
    for asynchronous invocation for a function, version, or alias.

    Update function's code, configuration, function's event invoke config This function compares data with new params
    and invokes update actions with appropriate arguments.

    Args:
        plan_state: idem --test state for update on AWS Lambda.
        name: The name of the Lambda function.
        update: a dictionary with newly passed values of params.
        function: response returned by describe on a lambda function
        code: a dictionary detailing existing code config.
        resource_id: The name/ ARN/ partial ARN of the AWS Lambda function, version, or alias.
        timeout(Dict, optional): Timeout configuration for create/update/deletion of AWS Lambda Function.
            * update (string) -- Timeout configuration for updating AWS Lambda Function
                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- Customized timeout configuration containing delay and max attempts.

    Returns:
        {"result": True|False, "comment": ("A tuple",), "ret": None}

    """
    result = dict(comment=(), result=True, ret=None)

    if not ctx.get("test", False):
        image_uri = None
        if code:
            image_uri = code.get("ImageUri")

        old_code_details = {
            "FunctionName": function.get("FunctionName"),
            "Architectures": function.get("Architectures"),
            "RevisionId": function.get("RevisionId"),
            "ImageUri": image_uri,
        }
        new_code_details = {
            "FunctionName": update.get("FunctionName"),
            "Architectures": update.get("Architectures"),
            "RevisionId": update.get("RevisionId"),
            "ImageUri": update.get("ImageUri"),
        }

        diff_in_code = differ.deep_diff(old_code_details, new_code_details)

        try:
            if diff_in_code.get("new"):
                publish = update.get("Publish", False)
                update_function_code = await hub.exec.boto3.client[
                    "lambda"
                ].update_function_code(
                    ctx=ctx,
                    FunctionName=resource_id,
                    ZipFile=update.get("ZipFile"),
                    S3Bucket=update.get("S3Bucket"),
                    S3Key=update.get("S3Key"),
                    S3ObjectVersion=update.get("S3ObjectVersion"),
                    ImageUri=diff_in_code.get("new").get("ImageUri"),
                    Publish=publish,
                    DryRun=False,
                    RevisionId=diff_in_code.get("new").get("RevisionId")
                    if diff_in_code.get("new").get("RevisionId")
                    else function.get("RevisionId"),
                    Architectures=diff_in_code.get("new").get("Architectures"),
                )
                result["result"] = result["result"] and update_function_code["result"]
                result["comment"] = result["comment"] + update_function_code["comment"]
                if not result["result"]:
                    return result

                waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                    default_delay=1,
                    default_max_attempts=20,
                    timeout_config=timeout.get("update") if timeout else None,
                )
                hub.log.debug(f"Waiting on updating aws.lambda.function '{name}'")
                try:
                    await hub.tool.boto3.client.wait(
                        ctx,
                        "lambda",
                        "function_updated",
                        FunctionName=resource_id,
                        WaiterConfig=waiter_config,
                    )
                except Exception as e:
                    result["comment"] = result["comment"] + (str(e),)
                    result["result"] = False

            update_function_config = await hub.exec.boto3.client[
                "lambda"
            ].update_function_configuration(
                ctx=ctx,
                FunctionName=resource_id,
                Role=update.get("Role"),
                Handler=update.get("Handler"),
                Description=update.get("Description"),
                Timeout=update.get("Timeout"),
                MemorySize=update.get("MemorySize"),
                VpcConfig=update.get("VpcConfig"),
                Environment=update.get("Environment"),
                Runtime=update.get("Runtime"),
                DeadLetterConfig=update.get("DeadLetterConfig"),
                KMSKeyArn=update.get("KMSKeyArn"),
                TracingConfig=update.get("TracingConfig"),
                RevisionId=update.get("RevisionId"),
                Layers=update.get("Layers"),
                FileSystemConfigs=update.get("FileSystemConfigs"),
                ImageConfig=update.get("ImageConfig"),
            )

            result["result"] = result["result"] and update_function_config.get("result")
            result["comment"] = result["comment"] + update_function_config["comment"]

            if not result["result"]:
                return result

            waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=1,
                default_max_attempts=20,
                timeout_config=timeout.get("update") if timeout else None,
            )
            hub.log.debug(f"Waiting on updating aws.lambda.function '{name}'")
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "lambda",
                    "function_updated",
                    FunctionName=resource_id,
                    WaiterConfig=waiter_config,
                )
            except Exception as e:
                result["comment"] = result["comment"] + (str(e),)
                result["result"] = False
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    else:
        update_params = OrderedDict(
            {
                "name": update.get("FunctionName"),
                "role": update.get("Role"),
                "code": update.get("Code"),
                "resource_id": update.get("FunctionName"),
                "runtime": update.get("Runtime"),
                "handler": update.get("Handler"),
                "description": update.get("Description"),
                "function_timeout": update.get("Timeout"),
                "memory_size": update.get("MemorySize"),
                "publish": update.get("Publish"),
                "vpc_config": update.get("VpcConfig"),
                "package_type": update.get("PackageType"),
                "dead_letter_config": update.get("DeadLetterConfig"),
                "environment": update.get("Environment"),
                "kms_key_arn": update.get("KMSKeyArn"),
                "tracing_config": update.get("TracingConfig"),
                "tags": update.get("Tags"),
                "layers": update.get("Layers"),
                "file_system_configs": update.get("FileSystemConfigs"),
                "image_config": update.get("ImageConfig"),
                "code_signing_config_arn": update.get("CodeSigningConfigArn"),
                "architectures": update.get("Architectures"),
                "qualifier": update.get("Qualifier"),
                "maximum_retry_attempts": update.get("MaximumRetryAttempts"),
                "maximum_event_age_in_seconds": update.get(
                    "maximum_event_age_in_seconds"
                ),
                "destination_config": update.get("DestinationConfig"),
                "revision_id": update.get("RevisionId"),
            }
        )
        for key, value in update_params.items():
            if value is None:
                if key not in plan_state:
                    plan_state[key] = None
            else:
                plan_state[key] = value
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.create_transit_gateway
hub.exec.boto3.client.ec2.delete_transit_gateway
hub.exec.boto3.client.ec2.describe_transit_gateways
hub.exec.boto3.client.ec2.modify_transit_gateway
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]
TREQ = {
    "absent": {
        "require": [
            "aws.ec2.transit_gateway_vpc_attachment.absent",
        ],
    },
}

__reconcile_wait__ = {"static": {"wait_in_seconds": 60}}


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    description: str = None,
    options: make_dataclass(
        "TransitGatewayRequestOptions",
        [
            ("AmazonSideAsn", int, field(default=None)),
            ("AutoAcceptSharedAttachments", str, field(default=None)),
            ("DefaultRouteTableAssociation", str, field(default=None)),
            ("DefaultRouteTablePropagation", str, field(default=None)),
            ("VpnEcmpSupport", str, field(default=None)),
            ("DnsSupport", str, field(default=None)),
            ("MulticastSupport", str, field(default=None)),
            ("TransitGatewayCidrBlocks", List[str], field(default=None)),
        ],
    ) = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a transit gateway. You can use a transit gateway to interconnect your virtual private clouds (VPC) and
    on-premises networks. After the transit gateway enters the available state, you can attach your VPCs and VPN
    connections to the transit gateway. To attach your VPCs, use CreateTransitGatewayVpcAttachment. To attach a VPN
    connection, use CreateCustomerGateway to create a customer gateway and specify the ID of the customer gateway
    and the ID of the transit gateway in a call to CreateVpnConnection. When you create a transit gateway, we create
    a default transit gateway route table and use it as the default association route table and the default
    propagation route table. You can use CreateTransitGatewayRouteTable to create additional transit gateway route
    tables. If you disable automatic route propagation, we do not create a default transit gateway route table. You
    can use EnableTransitGatewayRouteTablePropagation to propagate routes from a resource attachment to a transit
    gateway route table. If you disable automatic associations, you can use AssociateTransitGatewayRouteTable to
    associate a resource attachment with a transit gateway route table.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): AWS Transit gateway id.
        description(Text, Optional): description for the transit gateway to identify the resource
        options(Dict[str, Any], optional): The transit gateway options. Defaults to None.
            * AmazonSideAsn (int, optional): A private Autonomous System Number (ASN) for the Amazon side of a BGP session. The range is
                64512 to 65534 for 16-bit ASNs and 4200000000 to 4294967294 for 32-bit ASNs. The default is
                64512.
            * AutoAcceptSharedAttachments (str, optional): Enable or disable automatic acceptance of attachment requests. Disabled by default.
            * DefaultRouteTableAssociation (str, optional): Enable or disable automatic association with the default association route table. Enabled by
                default.
            * DefaultRouteTablePropagation (str, optional): Enable or disable automatic propagation of routes to the default propagation route table.
                Enabled by default.
            * VpnEcmpSupport (str, optional): Enable or disable Equal Cost Multipath Protocol support. Enabled by default.
            * DnsSupport (str, optional): Enable or disable DNS support. Enabled by default.
            * MulticastSupport (str, optional): Indicates whether multicast is enabled on the transit gateway
            * TransitGatewayCidrBlocks (List[str], optional): One or more IPv4 or IPv6 CIDR blocks for the transit gateway. Must be a size /24 CIDR block or
                larger for IPv4, or a size /64 CIDR block or larger for IPv6.
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the transit gateway.
            Each tag consists of a key name and an associated value. Defaults to None.
            * (Key, optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                characters. May not begin with aws:.
            * (Value, optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.

    Request Syntax:
        [transit-gateway-resource-name]:
          aws.ec2.transit_gateway.present:
            - resource_id: 'string'
            - description: 'string'
            - options:
                AmazonSideAsn: int
                AutoAcceptSharedAttachments: 'string'
                DefaultRouteTableAssociation: 'string'
                DefaultRouteTablePropagation: 'string'
                VpnEcmpSupport: 'string'
                DnsSupport: 'string'
                MulticastSupport: 'string'
                TransitGatewayCidrBlocks:
                  - 'string'
                  - 'string'
            - tags:
              - Key: 'string'
                Value: 'string'
    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            tgw-02994a8dda824c337:
                  aws.ec2.transit_gateway.present:
                  - resource_id: tgw-02994a8dda824c337
                  - description: test transit gateway
                  - options:
                      AmazonSideAsn: 64512
                      AutoAcceptSharedAttachments: enable
                      DefaultRouteTableAssociation: enable
                      DefaultRouteTablePropagation: enable
                      DnsSupport: enable
                      MulticastSupport: disable
                      TransitGatewayCidrBlocks:
                      - 10.0.0.0/24
                      - 198.0.0.0/16
                      VpnEcmpSupport: enable
                  - tags:
                    - Key: BU
                      Value: vRA-CS-GR
                    - Key: Organization
                      Value: vmw
                    - Key: Name
                      Value: transit-gateway-2

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    if resource_id:
        before = await hub.exec.aws.ec2.transit_gateway.get_transit_gateway_by_id(
            ctx, resource_id
        )
        if not before["result"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result
    try:
        if before and before["ret"]:
            result[
                "old_state"
            ] = hub.tool.aws.ec2.conversion_utils.convert_raw_transit_gateway_to_present(
                raw_resource=before["ret"], idem_resource_name=name
            )
            plan_state = copy.deepcopy(result["old_state"])
            current_status = result["old_state"]["state"]
            if current_status == "available":
                update_ret = await hub.exec.aws.ec2.transit_gateway.update_options(
                    ctx=ctx,
                    transit_gateway_id=result["old_state"].get("resource_id"),
                    old_description=result["old_state"].get("description"),
                    new_description=description,
                    old_options=result["old_state"].get("options"),
                    new_options=options,
                )
                result["comment"] = update_ret["comment"]
                result["result"] = update_ret["result"]
                resource_updated = resource_updated or bool(update_ret["ret"])

                if update_ret["ret"] is not None and ctx.get("test", False):
                    if update_ret["ret"].get("options") is not None:
                        plan_state["options"] = update_ret["ret"].get("options")
                    if update_ret["ret"].get("description") is not None:
                        plan_state["description"] = update_ret["ret"].get("description")

                hub.tool.aws.tag_utils.convert_tag_dict_to_list(
                    result["old_state"].get("tags")
                )

                if tags is not None and tags != result["old_state"].get("tags"):
                    # Update tags
                    update_ret = await hub.exec.aws.ec2.tag.update_tags(
                        ctx=ctx,
                        resource_id=result["old_state"].get("resource_id"),
                        old_tags=result["old_state"].get("tags"),
                        new_tags=tags,
                    )
                    result["comment"] = result["comment"] + update_ret["comment"]
                    result["result"] = result["result"] and update_ret["result"]
                    resource_updated = resource_updated or update_ret["result"]
                    if ctx.get("test", False) and update_ret["ret"] is not None:
                        plan_state["tags"] = update_ret["ret"]
            else:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"] = result["comment"] + (
                    f"aws.ec2.transit_gateway {name} is in {current_status} state",
                )
        else:
            if ctx.get("test", False):
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "description": description,
                        "options": options,
                        "tags": tags,
                    },
                )
                result["comment"] = result["comment"] + (
                    f"Would create aws.ec2.transit_gateway {name}",
                )
                return result
            ret = await hub.exec.boto3.client.ec2.create_transit_gateway(
                ctx,
                Description=description,
                Options=options,
                TagSpecifications=[
                    {
                        "ResourceType": "transit-gateway",
                        "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
                    }
                ]
                if tags
                else None,
            )

            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                return result

            resource_id = ret["ret"]["TransitGateway"]["TransitGatewayId"]
            result["comment"] = result["comment"] + (f"Created '{resource_id}'",)
    except hub.tool.boto3.exception.ClientError as e:
        result["result"] = False
        result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.aws.ec2.transit_gateway.get_transit_gateway_by_id(
                ctx, resource_id
            )
            result[
                "new_state"
            ] = hub.tool.aws.ec2.conversion_utils.convert_raw_transit_gateway_to_present(
                raw_resource=after["ret"], idem_resource_name=name
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified transit gateway.

    Args:
        name(Text): An Idem name of the transit gateway.
        resource_id(Text, optional): AWS Transit gateway id. Idem automatically considers this resource being absent
         if this field is not specified.

    Returns:
        Dict[str, Any]

    Request Syntax:
            [transit-gateway-resource-name]:
              aws.ec2.transit_gateway.absent:
                - name: 'string'
                - resource_id: 'string'


    Examples:

        .. code-block:: sls

            tgw-02994a8dda824c337:
              aws.ec2.transit_gateway.absent:
                - name: value
                - resource_id: tgw-02994a8dda824c337
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.transit_gateway", name=name
        )
        return result
    before = await hub.exec.aws.ec2.transit_gateway.get_transit_gateway_by_id(
        ctx, resource_id
    )
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result
    before = before["ret"]
    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.transit_gateway", name=name
        )
    elif before["State"] == "deleted":
        result["comment"] = (f"aws.ec2.transit_gateway '{name}' is in deleted state.",)
    else:
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_transit_gateway_to_present(
            raw_resource=before, idem_resource_name=name
        )
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.ec2.transit_gateway", name=name
            )
            return result
        try:
            ret = await hub.exec.boto3.client.ec2.delete_transit_gateway(
                ctx, TransitGatewayId=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                result["result"] = False
                return result
            after = await hub.exec.aws.ec2.transit_gateway.get_transit_gateway_by_id(
                ctx, resource_id
            )
            result[
                "new_state"
            ] = hub.tool.aws.ec2.conversion_utils.convert_raw_transit_gateway_to_present(
                raw_resource=after["ret"], idem_resource_name=name
            )
            result["comment"] = result["comment"] + (
                f"Deleted aws.ec2.transit_gateway '{name}'",
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Describes one or more transit gateways. By default, all transit gateways are described. Alternatively, you can
    filter the results.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ec2.transit_gateway
    """

    result = {}

    ret = await hub.exec.boto3.client.ec2.describe_transit_gateways(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe transit_gateway {ret['comment']}")
        return result
    for resource in ret["ret"]["TransitGateways"]:
        resource_id = resource.get("TransitGatewayId")
        resource_translated = (
            hub.tool.aws.ec2.conversion_utils.convert_raw_transit_gateway_to_present(
                raw_resource=resource, idem_resource_name=resource_id
            )
        )
        result[resource_id] = {
            "aws.ec2.transit_gateway.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result


def is_pending(hub, ret):
    new_state = ret.get("new_state", None)
    status = None
    resource_id = None
    if new_state:
        status = new_state.get("state")
        resource_id = new_state.get("resource_id")
    if status and isinstance(status, str):
        hub.log.debug(f"Transit gateway: {resource_id} is_pending() status {status}")
        if status.casefold() == "available" or status.casefold() == "deleted":
            hub.log.debug(
                f"No need to reconcile new state {new_state} for Transit gateway: {resource_id} with status {status}"
            )
            return False
        if (
            status.casefold() == "pending"
            or status.casefold() == "modifying"
            or status.casefold() == "deleting"
        ):
            hub.log.debug(
                f"Reconcile new state {new_state} for Transit gateway: {resource_id} with status {status}"
            )
            return True
    return (not ret["result"]) or bool(ret["changes"])



from typing import Any
from typing import Dict


async def get_resource_policy(
    hub,
    ctx,
    name: str,
) -> Dict[str, Any]:
    r"""
    Returns resource policy object for the given resource policy name
    Args:
        hub:
        ctx:
        name(Text): Name of the new policy. An Idem name of the resource

    Returns:
        Dict[str, Any]
    """
    result = dict(comment=(), result=True, ret=None)
    ret = await hub.exec.boto3.client.logs.describe_resource_policies(ctx)
    result["result"] = ret["result"]
    if not result["result"]:
        result["comment"] = result["comment"] + ret["comment"]
        return result
    if ret["ret"]:
        resource_policies = ret["ret"]["resourcePolicies"]
        for resource_policy in resource_policies:
            if resource_policy.get("policyName") == name:
                result["ret"] = resource_policy
                break
    return result



import copy
from typing import Any
from typing import Dict
from typing import List


async def update_tags(
    hub,
    ctx,
    resource_id: str,
    resource_type: str,
    old_tags: List[Dict[str, Any]] or Dict[str, Any] = None,
    new_tags: List[Dict[str, Any]] or Dict[str, Any] = None,
):
    """
    Update tags of AWS route53 hosted zone resources
    Args:
        hub:
        ctx:
        resource_id (Text): route53 hosted_zone resource id
        resource_type (Text): type of resource,
        old_tags (List): list of old tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
                         {tag-key: tag-value}
        new_tags (List): list of new tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
                         {tag-key: tag-value}

    Returns:
        {"result": True|False, "comment": ("A message",), "ret": Dict}

    """

    tags_to_add = {}
    tags_to_remove = {}
    if isinstance(old_tags, List):
        old_tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(old_tags)
    if isinstance(new_tags, List):
        new_tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(new_tags)
    if new_tags is not None:
        tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )
    result = dict(comment=(), result=True, ret={})
    if (not tags_to_remove) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result
    elif not tags_to_remove:
        tags_to_remove = None
    elif not tags_to_add:
        tags_to_add = None
    if not ctx.get("test", False):
        change_tag_resp = await hub.exec.boto3.client.route53.change_tags_for_resource(
            ctx,
            ResourceType=resource_type,
            ResourceId=resource_id,
            AddTags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags_to_add),
            RemoveTagKeys=list(tags_to_remove) if tags_to_remove else None,
        )
        if not change_tag_resp["result"]:
            response_message = change_tag_resp["comment"]
            hub.log.debug(
                f"Could not modify tags for {resource_id} with error {response_message}"
            )
            result["comment"] = (
                f"Could not modify tags for {resource_id} with error {response_message}",
            )
            result["result"] = False
            return result
        hub.log.debug(f"modified tags for {resource_id}")

    result["ret"] = new_tags
    result["comment"] = (f"Update tags: Add [{tags_to_add}] Remove [{tags_to_remove}]",)
    return result



import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    hosted_zone_name: str,
    caller_reference: str,
    resource_id: str = None,
    vpcs: make_dataclass(
        "VPC",
        [("VPCRegion", str, field(default=None)), ("VPCId", str, field(default=None))],
    ) = None,
    config: make_dataclass(
        "HostedZoneConfig",
        [
            ("Comment", str, field(default=None)),
            ("PrivateZone", bool, field(default=None)),
        ],
    ) = None,
    delegation_set_id: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass("Tag", [("Key", str), ("Value", str, field(default=None))])
    ] = None,
) -> Dict[str, Any]:
    r"""
        Creates a new public or private hosted zone. You create records in a public hosted zone to define how you want
        to route traffic on the internet for a domain, such as example.com, and its subdomains (apex.example.com,
        acme.example.com). You create records in a private hosted zone to define how you want to route traffic for a
        domain and its subdomains within one or more Amazon Virtual Private Clouds (Amazon VPCs).   You can't convert a
        public hosted zone to a private hosted zone or vice versa. Instead, you must create a new hosted zone with the
        same name and create new resource record sets.  For more information about charges for hosted zones, see Amazon
        Route 53 Pricing. Note the following:   You can't create a hosted zone for a top-level domain (TLD) such as
        .com.   For public hosted zones, Route 53 automatically creates a default SOA record and four NS records for the
        zone. For more information about SOA and NS records, see NS and SOA Records that Route 53 Creates for a Hosted
        Zone in the Amazon Route 53 Developer Guide. If you want to use the same name servers for multiple public hosted
        zones, you can optionally associate a reusable delegation set with the hosted zone. See the DelegationSetId
        element.   If your domain is registered with a registrar other than Route 53, you must update the name servers
        with your registrar to make Route 53 the DNS service for the domain. For more information, see Migrating DNS
        Service for an Existing Domain to Amazon Route 53 in the Amazon Route 53 Developer Guide.    When you submit a
        CreateHostedZone request, the initial status of the hosted zone is PENDING. For public hosted zones, this means
        that the NS and SOA records are not yet available on all Route 53 DNS servers. When the NS and SOA records are
        available, the status of the zone changes to INSYNC. The CreateHostedZone request requires the caller to have an
        ec2:DescribeVpcs permission.

        Args:
            name(Text): An Idem name of the hosted zone resource.
            hosted_zone_name(Text): A unique string that identifies a hosted zone.
            caller_reference(Text): A unique string that identifies the request and that allows failed CreateHostedZone
                                      requests to be retried without the risk of executing the operation twice.
                                      You must use a unique CallerReference string every time you submit a CreateHostedZone request.
                                      CallerReference can be any unique string, for example, a date/time stamp.,
            resource_id(Text, optional): AWS route53 hosted zone ID.
            config(Dict[str, Any], optional): (Optional) A complex type that contains the following optional values:   For public and private
                hosted zones, an optional comment   For private hosted zones, an optional PrivateZone element
                If you don't specify a comment or the PrivateZone element, omit HostedZoneConfig and the other
                elements. Defaults to None.
                * Comment (str, optional): Any comments that you want to include about the hosted zone.
                * PrivateZone (bool, optional): A value that indicates whether this is a private hosted zone.
            vpcs(Dict[str, Any], optional): (Private hosted zones only) A complex type that contains information about the Amazon VPC that
                you're associating with this hosted zone. You can specify only one Amazon VPC when you create a
                private hosted zone. If you are associating a VPC with a hosted zone with this request, the
                paramaters VPCId and VPCRegion are also required. To associate additional Amazon VPCs with the
                hosted zone, use AssociateVPCWithHostedZone after you create a hosted zone. Defaults to None.
                * VPCRegion (str, optional): (Private hosted zones only) The region that an Amazon VPC was created in.
                * VPCId (str, optional): (Private hosted zones only) The ID of an Amazon VPC.
            delegation_set_id(string, optional): Id of a reusable delegation set,
            tags(List or Dict, optional): list of tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
                {tag-key: tag-value} The tags to assign to the hosted zone. Defaults to None.
                * Key (Text) -- The key of the tag. Tag keys are case-sensitive and accept a maximum of 127 Unicode characters. May not begin with aws: .
                * Value (Text) -- The value of the tag. Tag values are case-sensitive and accept a maximum of 255 Unicode characters.


            Request Syntax:
            [hosted_zone_id]:
              aws.route53.hosted_zone.present:
              - name: 'string'
              - hosted_zone_name: 'string'
              - resource_id: 'string'
              - caller_reference: 'string'
              - config:
                  PrivateZone: 'string'
                  Comment: 'string'
              - vpcs:
                  VPCId: 'string'
                  VPCRegion: 'string'
              - delegation_set_id: 'string'
    -         - tags:
                - Key: 'string'
                  Value: 'string'

        Returns:
            Dict[str, Any]

        Examples:

            .. code-block:: sls

                resource_is_present:
                  aws.route53.hosted_zone.present:
                    - name: value
                    - hosted_zone_name: value
                    - resource_id: value
                    - caller_reference: value
                    - config:
                        PrivateZone: True
                        Comment: 'description of hosted zone config'
                    - vpcs:
                        VPCId: 'vpc id'
                        VPCRegion: 'us-east1'
                    - delegation_set_id: 'string'
    -               - tags:
                        - Key: 'tag key'
                          Value: 'tag value'
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False

    if resource_id:
        before = await hub.exec.boto3.client.route53.get_hosted_zone(
            ctx, Id=resource_id
        )

    if before:
        result["comment"] = (f"aws.route53.hostedzone '{name}' already exists.",)
        old_tags = None
        tags_ret = await hub.exec.boto3.client.route53.list_tags_for_resource(
            ctx, ResourceType="hostedzone", ResourceId=resource_id
        )
        if tags_ret["result"]:
            old_tags = tags_ret["ret"].get("ResourceTagSet").get("Tags", [])
        else:
            result["result"] = False
            result["comment"] = tags_ret["comment"]
            return result

        result[
            "old_state"
        ] = hub.tool.aws.route53.conversion_utils.convert_raw_hosted_zone_to_present(
            raw_resource=before,
            idem_resource_name=name,
            tags=old_tags,
        )
        plan_state = copy.deepcopy(result["old_state"])
        # Check if comments needs to be updated.
        if (
            config is not None
            and config.get("Comment")
            and config.get("Comment") != plan_state["config"].get("Comment")
        ):
            if ctx.get("test", False):
                result["comment"] = result["comment"] + (
                    f"Would update aws.route53.hostedzone '{name}'",
                )
                resource_updated = True
                plan_state["config"] = config
            else:
                update_ret_comment = (
                    await hub.exec.boto3.client.route53.update_hosted_zone_comment(
                        ctx,
                        Id=resource_id,
                        Comment=config.get("Comment"),
                    )
                )
                if update_ret_comment["result"]:
                    resource_updated = True
                    comment = config.get("Comment")
                    result["comment"] = result["comment"] + (
                        f"Updated comment to '{comment}'",
                    )
                else:
                    result["comment"] = update_ret_comment["comment"]
                    result["result"] = False

        # Check if tags needs to be updated
        if tags is not None:
            update_ret_tags = await hub.exec.aws.route53.tag.update_tags(
                ctx,
                resource_type="hostedzone",
                resource_id=resource_id,
                old_tags=old_tags,
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_ret_tags["comment"]
            if not update_ret_tags["result"]:
                result["result"] = False
                return result
            resource_updated = resource_updated or bool(update_ret_tags["ret"])
            if ctx.get("test", False) and update_ret_tags["ret"]:
                plan_state["tags"] = update_ret_tags["ret"]
    else:
        tags_dict = None
        if tags is not None:
            if isinstance(tags, List):
                tags_dict = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
            else:
                tags_dict = tags
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "hosted_zone_name": hosted_zone_name,
                    "caller_reference": caller_reference,
                    "config": config,
                    "vpcs": vpcs,
                    "delegation_set_id": delegation_set_id,
                    "tags": tags_dict,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.route53.hosted_zone", name=name
            )
            return result

        ret = await hub.exec.boto3.client.route53.create_hosted_zone(
            ctx,
            Name=hosted_zone_name,
            VPC=vpcs,
            CallerReference=caller_reference,
            HostedZoneConfig=config,
            DelegationSetId=delegation_set_id,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.route53.hosted_zone", name=name
        )
        resource_id = ret["ret"]["HostedZone"]["Id"].split("/")[-1]
        if tags is not None:
            # add tag
            add_tags = await hub.exec.aws.route53.tag.update_tags(
                ctx,
                resource_type="hostedzone",
                resource_id=resource_id,
                old_tags=None,
                new_tags=tags_dict,
            )
            if not add_tags["result"]:
                error = add_tags["comment"]
                result["result"] = False
                hub.log.debug(f"Tag addition failed for {name} with error {error}")
                result["comment"] = result["comment"] + add_tags["comment"]

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif (not before) or resource_updated:
        after = await hub.exec.boto3.client.route53.get_hosted_zone(ctx, Id=resource_id)
        if after["result"]:
            tags_ret = await hub.exec.boto3.client.route53.list_tags_for_resource(
                ctx,
                ResourceType="hostedzone",
                ResourceId=resource_id,
            )
            if tags_ret["result"]:
                tags = tags_ret["ret"]["ResourceTagSet"].get("Tags")
            result[
                "new_state"
            ] = hub.tool.aws.route53.conversion_utils.convert_raw_hosted_zone_to_present(
                raw_resource=after,
                idem_resource_name=name,
                tags=tags,
            )
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes a hosted zone. If the hosted zone was created by another service, such as Cloud Map, see Deleting Public
    Hosted Zones That Were Created by Another Service in the Amazon Route 53 Developer Guide for information about
    how to delete it. (The process is the same for public and private hosted zones that were created by another
    service.) If you want to keep your domain registration but you want to stop routing internet traffic to your
    website or web application, we recommend that you delete resource record sets in the hosted zone instead of
    deleting the hosted zone.  If you delete a hosted zone, you can't undelete it. You must create a new hosted zone
    and update the name servers for your domain registration, which can require up to 48 hours to take effect. (If
    you delegated responsibility for a subdomain to a hosted zone and you delete the child hosted zone, you must
    update the name servers in the parent hosted zone.) In addition, if you delete a hosted zone, someone could
    hijack the domain and route traffic to their own resources using your domain name.  If you want to avoid the
    monthly charge for the hosted zone, you can transfer DNS service for the domain to a free DNS service. When you
    transfer DNS service, you have to update the name servers for the domain registration. If the domain is
    registered with Route 53, see UpdateDomainNameservers for information about how to replace Route 53 name servers
    with name servers for the new DNS service. If the domain is registered with another registrar, use the method
    provided by the registrar to update name servers for the domain registration. For more information, perform an
    internet search on "free DNS service." You can delete a hosted zone only if it contains only the default SOA
    record and NS resource record sets. If the hosted zone contains other resource record sets, you must delete them
    before you can delete the hosted zone. If you try to delete a hosted zone that contains other resource record
    sets, the request fails, and Route 53 returns a HostedZoneNotEmpty error. For information about deleting records
    from your hosted zone, see ChangeResourceRecordSets. To verify that the hosted zone has been deleted, do one of
    the following:   Use the GetHostedZone action to request information about the hosted zone.   Use the
    ListHostedZones action to get a list of the hosted zones associated with the current Amazon Web Services
    account.

    Args:
        name(Text): An Idem name of the hosted zone resource.
        resource_id(Text, optional): AWS route53 hosted zone ID. Idem automatically considers this resource being absent
         if this field is not specified.

    Request Syntax:
        [hosted-zone-name]:
          aws.route53.hosted_zone.present:
          - name: 'string'
          - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.route53.hosted_zone.absent:
                - name: value
                - resource_id: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.route53.hosted_zone", name=name
        )
        return result
    before = await hub.exec.boto3.client.route53.get_hosted_zone(ctx, Id=resource_id)

    if not before["result"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.route53.hosted_zone", name=name
        )
    else:
        tags = None
        tags_ret = await hub.exec.boto3.client.route53.list_tags_for_resource(
            ctx, ResourceType="hostedzone", ResourceId=resource_id
        )
        if tags_ret["result"]:
            tags = tags_ret["ret"]["ResourceTagSet"].get("Tags")
        result[
            "old_state"
        ] = hub.tool.aws.route53.conversion_utils.convert_raw_hosted_zone_to_present(
            raw_resource=before,
            idem_resource_name=name,
            tags=tags,
        )
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.route53.hosted_zone", name=name
            )
            return result
        ret = await hub.exec.boto3.client.route53.delete_hosted_zone(
            ctx, Id=resource_id
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.route53.hosted_zone", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Retrieves a list of the public and private hosted zones that are associated with the current Amazon Web Services
    account.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.route53.hosted_zone
    """

    result = {}

    ret = await hub.exec.boto3.client.route53.list_hosted_zones(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe hosted_zone {ret['comment']}")
        return result

    for hosted_zone in ret["ret"]["HostedZones"]:
        tags = None
        resource_id = hosted_zone.get("Id")
        resource_id = resource_id.split("/")[-1]
        individual_hosted_zone = await hub.exec.boto3.client.route53.get_hosted_zone(
            ctx, Id=resource_id
        )

        tags_ret = await hub.exec.boto3.client.route53.list_tags_for_resource(
            ctx, ResourceType="hostedzone", ResourceId=resource_id
        )
        if tags_ret["result"]:
            tags = tags_ret["ret"]["ResourceTagSet"].get("Tags")

        resource_translated = (
            hub.tool.aws.route53.conversion_utils.convert_raw_hosted_zone_to_present(
                raw_resource=individual_hosted_zone,
                idem_resource_name=resource_id,
                tags=tags,
            )
        )
        result[resource_id] = {
            "aws.route53.hosted_zone.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result


async def search(
    hub,
    ctx,
    name: str,
    hosted_zone_name: str = None,
    resource_id: str = None,
    private_zone: bool = None,
    vpc_id: str = None,
    tags: List = None,
):
    """
    Use an un-managed hosted_zone as a data-source. Supply one or more of the inputs as the filter.

    Args:
        name(string): The name of the Idem state.
        hosted_zone_name(string, optional): The domain name of hosted_zone
        resource_id(string, optional): AWS hosted_zone id to filter the resource.If resource_id is mentioned then we
                                       directly return the resource without checking other filters
        private_zone(bool, optional): Bool argument to specify a private hosted_zone. One of the filter option for hosted_zone
        vpc_id(string, optional):The vpc_id associated with the hosted_zone.One of the filter option for hosted_zone
        tags(List, optional): Tags of the hosted_zone. One of the filter option for hosted_zone


    Request Syntax:
        [Idem-state-name]:
          aws.route53.hosted_zone.search:
          - hosted_zone_name: 'string'
          - resource_id: 'string'
          - private_zone: 'bool'
          - vpc_id: 'string'
          - tags: 'List'

    Returns:
        {"comment": A message tuple, "old_state": Dict, "new_state": Dict, "name": string, "result": bool}

        Examples:

            my-unmanaged-hosted_zone:
              aws.route53.hosted_zone.search:
                - hosted_zone_name: idem-test-example-86a1d12e-4d68-4979-96ff-8608c33e4b9f.com.
                - resource_id: Z02905093MT09M4X5AHR5
                - private_zone: false
                - vpc_id: vpc-0e87ce8f2c3317d53
                - tags:
                  - Key: Name
                    Value: idem-test-hosted_zone

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if resource_id:
        # If resource_id is mentioned directly returning the hosted_zone
        ret = await hub.exec.boto3.client.route53.get_hosted_zone(ctx, Id=resource_id)
        if not ret["result"]:
            result["result"] = ret["result"]
            result["comment"] = (
                f"Could not find hosted_zone with given {resource_id} resource_id {ret['comment']}",
            )
            return result
        result["comment"] = (
            f"Found hosted_zone with given resource_id {resource_id}, other filters will not be considered",
        )
        hosted_zone = ret
        tags_ret = await hub.exec.boto3.client.route53.list_tags_for_resource(
            ctx, ResourceType="hostedzone", ResourceId=resource_id
        )
        if not tags_ret["result"]:
            result["comment"] = tags_ret["comment"]
            result["result"] = False
            return result
        tags = tags_ret["ret"]["ResourceTagSet"].get("Tags")
    else:
        # Get all the hosted_zones to apply the filters
        describe_ret = await hub.exec.aws.route53.hosted_zone.get_all_hosted_zones(ctx)
        if not describe_ret["result"]:
            result["result"] = False
            result["comment"] = (
                f"Could not describe all the aws.route53.hosted_zone resources to apply the filters, {describe_ret['comment']}",
            )
            return result

        ret = hub.tool.aws.route53.hosted_zone_utils.get_hosted_zone_with_filters(
            raw_hosted_zones=describe_ret["ret"],
            hosted_zone_name=hosted_zone_name,
            private_zone=private_zone,
            vpc_id=vpc_id,
            tags=tags,
        )
        result["comment"] = result["comment"] + ret["comment"]
        if not ret["ret"]:
            result["result"] = ret["result"]
            return result

        hosted_zone = ret["ret"]
        tags = hosted_zone.get("tags")

    result[
        "old_state"
    ] = hub.tool.aws.route53.conversion_utils.convert_raw_hosted_zone_to_present(
        raw_resource=hosted_zone,
        idem_resource_name=name,
        tags=tags,
    )
    # Populate both "old_state" and "new_state" with the same data
    result["new_state"] = copy.deepcopy(result["old_state"])

    return result



DESCRIBE_FUNCTIONS = ("describe", "get", "search", "list")

DELETE_FUNCTIONS = (
    "delete",
    "disassociate",
    "reject",
    "deallocate",
    "unassign",
    "deregister",
    "deprovision",
    "revoke",
    "release",
    "terminate",
    "cancel",
    "disable",
)

CREATE_FUNCTIONS = (
    "create",
    "associate",
    "accept",
    "allocate",
    "assign",
    "register",
    "provision",
    "authorize",
    "run",
    "enable",
    "upload",
    "put",
    "publish",
    "request",
    "put",
)

NAME_PARAMETER = {
    "default": None,
    "doc": "An Idem name of the resource",
    "param_type": "Text",
    "required": True,
    "target": "hardcoded",
    "target_type": "arg",
}

RESOURCE_ID_PARAMETER = {
    "default": None,
    "doc": "An identifier of the resource in the provider",
    "param_type": "Text",
    "required": False,
    "target": "hardcoded",
    "target_type": "arg",
}

PRESENT_REQUEST_FORMAT = r"""
    # TODO: result["old_state"] and result["new_state"] should be populated with the same parameters as the present() input parameters.
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if resource_id:
        {{ function.hardcoded.resource_function_call }}
        before = {{ function.hardcoded.describe_function_call }}
    if before:
        # TODO perform day-2 modifications as needed here
        # TODO if ctx.test is True, no update call should be made to AWS, but parameter values should still be compared
        # and the "new_state" return should reflect the updated value that would be made if actual update calls happened.
        resource_updated = True
        if ctx.get("test", False) and resource_updated:
            # TODO populate result["new_state"]
            result["comment"] = (f"Would update aws.{{ function.hardcoded.service_name }}.{{ function.hardcoded.resource }} '{name}'",)
            return result
        if not resource_updated:
            result["comment"] = (f"'{name}' already exists",)
    else:
        if ctx.get("test", False):
            # TODO populate result["new_state"]
            result["comment"] = (f"Would create aws.{{ function.hardcoded.service_name }}.{{ function.hardcoded.resource }} {name}",)
            return result
        try:
            ret = await {{ function.hardcoded.create_function }}(
                ctx,
                {{ "ClientToken=name," if function.hardcoded.has_client_token }}
                **{{ parameter.mapping.kwargs|default({}) }}
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = (f"Created '{name}'",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    {{ function.hardcoded.waiter_call }}
    # TODO perform other modifications as needed here
    ...

    try:
        if ctx.get("test", False):
            # TODO populate result["new_state"]
            ...
        elif (not before) or resource_updated:
            after = {{ function.hardcoded.describe_function_call }}
            result["new_state"] = after
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result
"""

ABSENT_REQUEST_FORMAT = r"""
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    {{ function.hardcoded.resource_function_call }}
    before = {{ function.hardcoded.describe_function_call }}

    if not before:
        result["comment"] = (f"'{name}' already absent",)
    elif ctx.get("test", False):
        # TODO populate result["new_state"]
        result["comment"] = (f"Would delete aws.{{ function.hardcoded.service_name }}.{{ function.hardcoded.resource }} '{name}'",)
        return result
    else:
        # TODO populate result["old_state"]
        result["old_state"] = before
        try:
            ret = await {{ function.hardcoded.delete_function }}(
                ctx,
                {{ "ClientToken=name," if function.hardcoded.has_client_token }}
                **{{ parameter.mapping.kwargs|default({}) }}
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = (f"Deleted '{name}'",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    {{ function.hardcoded.waiter_call }}

    return result
"""

DESCRIBE_REQUEST_FORMAT = r"""
    result = {}
    ret = await {{ function.hardcoded.list_function}}(ctx)

    if not ret["status"]:
        hub.log.debug(f"Could not describe {{ function.hardcoded.resource }} {ret['comment']}")
        return {}

    # TODO: The parameters of the describe() output should be the same as the present() input parameters.
    for {{ function.hardcoded.resource }} in ret["ret"]["{{ function.hardcoded.list_item }}"]:
        # Including fields to match the 'present' function parameters
        # TODO convert the dictionary values from string to object by removing the quotes.
        # TODO From 'resource[param]' to resource[param]
        new_{{ function.hardcoded.resource }} = {{ function.hardcoded.present_params }}
        result[{{ function.hardcoded.resource }}["{{ function.hardcoded.resource_id }}"]] = {"{{ function.ref }}.present": new_{{ function.hardcoded.resource }}}

    return result
"""
